###############################################################################
## TimeTravel by Zoomj [0.1]
##
## File: timetravel.py
## Description: main plugin file
## Last modified: 03/07/2011

from __future__ import with_statement
from helpers import *
import gamethread
import es

_max_states = 600  ## Maximum states saved per player recording
_plugin = None     ## Global plugin singleton instance

###############################################################################
## TimeTravel player class

class TTPlayer(object):
    
    """ [Constructor] Set initial state """
    def __init__(self, userid):
        self.userid = userid
        self.paused = True
        self.tick = 0
        self.current_tick = 0
        self.state = []
        self.initial_play = False
        self.consecutive_fastforwards = 0
        self.consecutive_rewinds = 0
            
    """ Called from OnGameFrame - update client data per frame """
    def update(self):
        if self.paused: return
        else:
            if len(self.state) > _max_states:
                self.state.pop(0)
                self.tick -= 1
                self.current_tick -= 1
            self.tick += 1
            self.current_tick += 1
            self.state.append(self.tick)
            self.state[self.tick-1] = {
                'velocity' : get_player_velocity(self.userid),
                'angles' : get_player_angles(self.userid),
                'location' : get_player_location(self.userid),
            }
            
    """ Record client data from this point """
    def play(self):
        self.consecutive_fastforwards = 0
        self.consecutive_rewinds = 0
        if self.paused:
            if self.current_tick < self.tick:
                for x in xrange(self.current_tick, self.tick-1):
                    self.state.pop(-1)
                self.tick = self.current_tick
            if self.initial_play:
                obj = self.state[self.current_tick-1]
                set_player_velocity(self.userid, obj['velocity'])
            else: self.initial_play = True
            self.paused = False
            unfreeze_player(self.userid)
            
    """ Pause recording """
    def pause(self):
        self.paused = True
        freeze_player(self.userid)
        kill_player_velocity(self.userid)
        obj = self.state[self.current_tick-1]
        set_player_angles(self.userid, obj['angles'])
        set_player_location(self.userid, obj['location'])
        
    """ Rewind client recording by 0.5 seconds """
    def rewind(self):
        self.consecutive_fastforwards = 0
        self.consecutive_rewinds += 1
        rewind_tick_rate = 1
        if self.consecutive_rewinds >= 4:
            rewind_tick_rate = 5
        rewind_tick = (self.current_tick -1) - rewind_tick_rate
        if rewind_tick >= 0:
            self.current_tick = rewind_tick
            self.pause()
        
    """ Fast forward client recording by 0.5 seconds """
    def fastforward(self):
        self.consecutive_rewinds = 0
        self.consecutive_fastforwards += 1
        ff_tick_rate = 1
        if self.consecutive_fastforwards >= 4:
            ff_tick_rate = 5
        ff_tick = self.current_tick + ff_tick_rate
        if ff_tick < len(self.state):
            self.current_tick = ff_tick
            self.pause()
     
###############################################################################
## TimeTravel plugin class (manager)
     
class TTPlugin(object):
    
    """ [Constructor] Handle load event """
    def __init__(self):
        self.register_commands()
        self.clients = {}
        gamethread.delayedname(0.1, 'tt_tick', TTCallbacks.timer_tick)
        
    """ Register commands used by this plugin with cmdlib """
    def register_commands(self):
        regcmd_c('tt_play', TTCallbacks.cmd_play)
        regcmd_c('tt_stop', TTCallbacks.cmd_stop)
        regcmd_c('tt_pause', TTCallbacks.cmd_pause)
        regcmd_c('tt_rewind', TTCallbacks.cmd_rewind)
        regcmd_c('tt_fastforward', TTCallbacks.cmd_fastforward)
        
    """ Unregister commands used by this plugin with cmdlib """
    def unregister_commands(self):
        unregcmd_c('tt_play')
        unregcmd_c('tt_stop')
        unregcmd_c('tt_pause')
        unregcmd_c('tt_rewind')
        unregcmd_c('tt_fastforward')
        
    """ Return True if client data is being recorded """
    def has_client(self, userid):
        if self.clients.has_key(userid):
            return True
        else: return False
        
    """ Return TTPlayer client object if client data is being recorded """
    def get_client(self, userid):
        if self.clients.has_key(userid):
            return self.clients[userid]
        else: return None
        
    """ Start recording client data """
    def add_client(self, userid):
        if not self.clients.has_key(userid):
            self.clients[userid] = TTPlayer(userid)
            
    """ Stop recording client data and erase all existing data """
    def remove_client(self, userid):
        if self.clients.has_key(userid):
            del self.clients[userid]
            
    """ Update all clients whose data is being recorded """
    def update_clients(self):
        for userid, client in self.clients.items():
            if not userid in es.getUseridList():
                self.remove_client(userid)
            else:
                client.update()
        
    """ Destroy all client data """
    def destroy_clients(self):
        self.clients = {}
        
    """ [Destructor] Handle unload event """
    def nice_boat(self):
        gamethread.cancelDelayed('tt_tick')
        self.unregister_commands()

###############################################################################
## TimeTravel callback class
        
class TTCallbacks(object):

    """ Command callback for 'tt_play' """
    @staticmethod
    def cmd_play(userid, args):
        global _plugin
        if not _plugin.has_client(userid):
            _plugin.add_client(userid)
        client = _plugin.get_client(userid)
        if client:
            client.play()
            send_hint(userid, 'Recording game state')
            
    """ Command callback for 'tt_stop' """
    @staticmethod
    def cmd_stop(userid, args):
        global _plugin
        if _plugin.has_client(userid):
            _plugin.remove_client(userid)
            unfreeze_player(userid)
            send_hint(userid, 'Stopped recording game state')

    """ Command callback for 'tt_pause' """
    @staticmethod
    def cmd_pause(userid, args):
        global _plugin
        if _plugin.has_client(userid):
            client = _plugin.get_client(userid)
            if client:
                client.pause()
                send_hint(userid, 'State recording has been paused')
        
    """ Command callback for 'tt_rewind' """
    @staticmethod
    def cmd_rewind(userid, args):
        global _plugin
        if _plugin.has_client(userid):
            client = _plugin.get_client(userid)
            if client:
                client.rewind()
                send_hint(userid, 'State has been re-winded')
        
    """ Command callback for 'tt_fastforward' """
    @staticmethod
    def cmd_fastforward(userid, args):
        global _plugin
        if _plugin.has_client(userid):
            client = _plugin.get_client(userid)
            if client:
                client.fastforward()
                send_hint(userid, 'State has been fast-forwarded')
        
    """ Gamethread callback for 'tt_tick' """
    @staticmethod
    def timer_tick():
        global _plugin
        _plugin.update_clients()
        gamethread.delayedname(0.1, 'tt_tick', TTCallbacks.timer_tick)

###############################################################################
## EventScripts events
  
""" Load plugin """  
def load():
    global _plugin
    _plugin = TTPlugin()
    
""" Unload plugin """
def unload():
    global _plugin
    _plugin.nice_boat()
    
""" Remove client from plugin data when they disconnect """
def player_disconnect(e):
    global _plugin
    _plugin.remove_client(int(e['userid']))
    
""" Remove client from plugin data when they die """
def player_death(e):
    global _plugin
    _plugin.remove_client(int(e['userid']))
    
""" Remove client from plugin data when they (re-)spawn """
def player_spawn(e):
    global _plugin
    _plugin.remove_client(int(e['userid']))
    
""" Remove all clients from plugin data when the map changes """
def es_map_start(e):
    global _plugin
    _plugin.destroy_clients()

###############################################################################